/***********************************************************************

  CWEED Forth

Author: Rod Crawford

Date: 16/11/89
Mods:  9/ 4/90 Rod changed cweed interpret to support execution token
compilation rather than just straight CFA's
12/ 4/90 Rod Added OS9 Support
23/ 7/90 SFP Version number changed to 3.0

This file contains the CWeed Forth Interpreter, Compiler and Initialisation
Functions.

 ************************************************************************/



#include "cmodel.h"
#include "pinc.h"
#include "vmsize.h"
#include "lowram.h"
#include "nfmasks.h"
#include "strsizes.h"
#include "misc.h"

/**************************  external data *******************************/


#if IMPLEMENTATION == MicrosoftC
extern unsigned char huge ram[ RAM_TOP ];
#else
#if IMPLEMENTATION == OS9
extern remote unsigned char ram[ RAM_TOP ];
#else
extern unsigned char ram[ RAM_TOP ];
#endif
#endif

extern unsigned CELL sp, rp, dp;
extern unsigned CELL compiling;
extern int kill_switch;

extern struct from_fileblock {
    char filename[ FROM_FILENAME_LENGTH ];
    int page_number, line_number, line_in_page_number,
        last_page_required;
    FILE *stream;
};

extern struct from_fileblock output;


/******************* External Function Prototypes  ************************/

#ifdef KNR
extern unsigned CELL defined();
extern void do_error();
extern unsigned CELL lfa_to_nfa();
extern unsigned CELL nfa_to_cfa();
extern unsigned CELL comma_word();
extern unsigned CELL defined();
extern void compile_literal();
extern void execute();
extern void prompt();
extern int *get_line();
extern void init_io();
extern void strupper();
extern void strlower();
extern int number();
extern int is_immediate();
extern unsigned CELL execution_token();

#ifdef CHECK_STACKS
extern void push_data();
#else
#include "stacks.h"
#endif

#else
extern unsigned CELL defined(char *word_name);
extern void do_error(unsigned CELL error_number);
extern unsigned CELL lfa_to_nfa(unsigned CELL address);
extern unsigned CELL nfa_to_cfa(unsigned CELL nfa);
extern unsigned CELL comma_word(unsigned CELL data);
extern unsigned CELL defined(char *word_name);
extern void compile_literal(unsigned CELL value);
extern void execute(unsigned CELL cfa);
extern void prompt(void );
extern int get_line(void );
extern void init_io(void );
extern void strupper( char *string );
extern void strlower( char *string );
extern int number(char *string,int string_ptr,unsigned CELL *value);
extern int is_immediate(unsigned CELL nfa);
extern unsigned CELL execution_token(unsigned CELL lfa);

#ifdef CHECK_STACKS
extern void push_data(unsigned CELL data);
#else
#include "stacks.h"
#endif

#endif

#include "extproto.h"

/*************************  Function Prototypes ********************/
#ifdef KNR
void init_user_variables();
void cold_start();
void quit();
void cweed_interpret();
int compile_word();
void undefined();
int execute_word();
#else
void init_user_variables(void );
void cold_start(void );
void quit(void );
void cweed_interpret(void );
int compile_word(char *word_name);
void undefined(char *string);
int execute_word(char *word_name);
#endif



/*********************  init_user_variables  ***************************

Desc:  Initialises Required user variables to sensible values.

Alg:    The initial values that RP and SP should be set to are placed
in the user variable R0 and S0. N.B. These values are
0 and DATA_STACK_TOP respectively, since the return stack
grows up and the data stack grows down.

Mods This has been changed (14/07/90) so that both stacks grow 
upwards, starting at 0. This change was included for the sake
of consistency, comprehensibility, and compatibility with
most hardware stack machines.

 ************************************************************************/

void init_user_variables( void )
{
    unsigned CELL *cell;

    cell  = (unsigned CELL *) &ram[ R0 ];
    *cell = (CELL) -1;
    cell  = (unsigned CELL *) &ram[ S0 ];
    *cell = (CELL) -1;
    cell  = (unsigned CELL *) &ram[ BASE ];
    *cell = (CELL) 10;
}




/*******************************  cold_start  ****************************

Desc: This function is called when CWeed Forth is being started or
restarted during a bootstrap. It is not indended to be used
by full blown Pinc.

Alg:  Signs on intialises the kill switch and notes that we are not
compiling. Order is printed out and we execute quit.

 *************************************************************************/

void cold_start( void )
{
    fprintf( output.stream, "CWeed Forth 3.0\n\n" );
    compiling=0;
    kill_switch=0;

    order();
    quit();
}



/*****************************  quit  ********************************

Desc: CWeed Forth version of quit.

Alg:  The data stack is cleared.
While not at end of input stream
Clear the return stack and get a line of input.
If not end of input
Interpet the line and prompt accordingly.

 ****************************************************************/

void quit( void )
{
    int end_of_input = 0;

    fprintf( output.stream, "\n" );

    sp = (CELL) -1;
    while( !end_of_input ) {
        rp = (CELL) -1;

        end_of_input = (EOF == (int) get_line());
        if( !end_of_input ) {
            cweed_interpret();
            if( !compiling ) prompt();
        }
    }
}



/***********************  cweed_interpret   ***************************

Desc: CWeed Forths Version of interpret.

Alg: While there is a word in the input stream
if we are compiling try and compile it
else try and execute it.

 ************************************************************************/

void cweed_interpret()
{
    char word[ WORD_LENGTH ];
    unsigned CELL lfa;
    CELL value;

    while( get_word( word )) {
        if( !compiling ) execute_word( word );
        else compile_word( word );
    }
}



/***************************  compile_word  ***************************

Desc: Searches for the defintiion of the given word in the dictionary.
If it cannot be found
If the word_name can be the representation of a literal number
compile that literal
Else note the word is undefined
Else If the word is not immediate
compile the cfa of the word
Else
execute the word.

Note how we turn off the compiler temporarily while searching
for the execution token of an immediate word. This
causes execution_token to return the cfa of the word
not the contents of the cfa. Thus the cfa can executed,
(the contents if they were an opcode could not).

In:   word_name  string representing either the name of a word to be
compiled, a number or garbage.

 **********************************************************************/

compile_word( word_name )
    char *word_name;
{
    unsigned CELL lfa, token;
    CELL value;

    //    if( !( lfa = defined_word( word_name,(unsigned char)*word_name )))
    if( !( lfa = defined_word( word_name)))
        if( !number( word_name, 0, &value))  undefined( word_name );
        else compile_literal( value );
    else
        if( is_immediate( lfa_to_nfa( lfa ))) {
            compiling = FALSE;
            token = execution_token( lfa );
            compiling = TRUE;
            execute( token );
        }
        else
            dp = comma_word( execution_token( lfa ));

}





/***************************** undefined  *************************

Desc: Informs the user that the given string is undefined.

In:   string  the string that is undefined

 ******************************************************************/

void undefined( string )
    char string[];
{
    do_error( (CELL) 0 );
    compiling = 0;
}



/****************************  execute_word  **************************

Desc: Attempts to execute the word whose name is given. If the word
cannot be found then if the word name could represent a number
then that number is pushed on the stack.
If word name is none of the above the user is notified.

In:   word_name  string representing the word to be excuted, a number
or garbage

 *********************************************************************/

execute_word( word_name )
    char *word_name;
{
    unsigned CELL lfa;
    CELL value;

    if( !( lfa = defined_word( word_name ))) {
        if( !number( word_name, 0, &value)) undefined( word_name );
        else push_data( (unsigned CELL) value );
    }
    else {
        execute( execution_token( lfa ));
    }
}
