/*****************************  IO.C
 ***********************************

Author: Rod Crawford

Date: 16/11/89
Mods:  5/ 4/90 Rod Added support for small model in prompt() Output
12/ 4/90 Rod Added support for OS9

This file contains the Pinc IO manager

 *****************************************************************
 *********/

#include "cmodel.h"
#include "pinc.h"
#include "vmsize.h"
#include "lowram.h"
#include "strsizes.h"

#if IMPLEMENTATION == Linux
#include <termios.h>
#endif


/**************************  external data
 *******************************/

#if IMPLEMENTATION == MicrosoftC
extern unsigned char huge ram[ RAM_TOP ];
#else
#if IMPLEMENTATION == OS9
extern remote unsigned char ram[ RAM_TOP ];
#else
extern unsigned char ram[ RAM_TOP ];
#endif
#endif


extern unsigned CELL sp;



/**********************  PUBLIC GLOBAL DATA
 ****************************

 Input/Output is directed through the from_fileblock variables
 input, output and
 stderr. input and output default to stdin and stdout and can be
 redirected
 at a later during program execution. Under unix the terminal
 configuration
 is held in a termio struct. The last_tty_config varaible hold the
 tty config prior to it being set into raw mode the can then be used
 to restore the balance of things when required.

 *****************************************************************
 *******/
/* ag-special. The following is commented out for compatibilty
   with SunOS 3.x. Can be uncommented for SUNOS 4.0
   to provide proper key? etc.
   */

#if IMPLEMENTATION == SUN
struct termio last_tty_config;
#elif IMPLEMENTATION == Linux
struct termios oldt, newt;
int oldf;
#endif


/**********************  PRIVATE GLOBAL DATA
 ***************************

 Include files may be nested FROM_STACK_LENGTH-1 deep. The
 open streams of files which are being included are kept in
 from_fileblocks
 on from_file_stack. The stack is initialised to empty with stdin
 as
 output (the current file we are includeing from.
 All input takes place through the file descriptor
 input. Its definition follows below.

 The error handler prints out the appropriate error message
 from the error file. However if the error file is not found at
 initialisation time, the error message number only is printed.
 The fact of wether an error file was found is determined by
 wether the error file stream is NULL or not.

 *****************************************************************
 ********/

FILE *error_message_stream;
int from_stack_ptr = -1;

struct from_fileblock {
    char filename[ FROM_FILENAME_LENGTH ];
    int page_number, line_number, line_in_page_number,
        last_page_required;
    FILE *stream;
} output, input,
    from_file_stack[ FROM_FILE_STACK_LENGTH ];




/************************  FUNCTION PROTOTYPES
 **************************/

#ifdef KNR
void init_io();
void init_from_stack();
void push_from_fileblock();
struct from_fileblock pop_from_fileblock();
void trash_rest_of_line();
int get_string();
int get_word();
int get_line();
int get_basic_line();
void prompt();
void init_tty_config();
void set_raw_io();
void set_cooked_io();
int key_pressed();
struct from_fileblock make_from_fileblock();
void goto_page();
#else
void init_io(void );

void init_from_stack(void );
void push_from_fileblock(struct from_fileblock a_from_fileblock);
struct from_fileblock pop_from_fileblock(void );
void trash_rest_of_line(void );
int get_string(char *string,char delimiter);
int get_word(char *word);
int get_line(void );
int get_basic_line( char line[], int length, FILE *stream, int
        *page_break );
void prompt(void );
struct from_fileblock make_from_fileblock(char *filename, FILE
        *stream);
void goto_page( unsigned CELL page_number );
#endif



/****************************  init_io
 **********************************

Desc: All file streams barring stdin and stdout are closed.
input and output come from the standard streams.
Initialises Pincs IO. The line pointer >IN is set to 0.
The tty_configuration is setup, then the from stack
and the error handler are initialised.

 *****************************************************************
 ********/

void init_io(void )
{
    unsigned CELL *lineptr;

#if IMPLEMENTATION == MicrosoftC
    fcloseall();
#endif
    input = make_from_fileblock( "Console", stdin );
    output = make_from_fileblock( "Console", stdout );

    lineptr = (unsigned CELL *) &ram[ TO_IN ];
    *lineptr = (CELL) 0;
#if IMPLEMENTATION == SUN
    init_tty_config();
#endif
    init_from_stack();
    init_error_handler();
}


/***************************  init_from_stack
 ***************************

Desc: The from stack pointer is initialised to point one below the
last
item on the the from stack and the standard input stream is
pushed
onto the stack.

 *****************************************************************
 *********/

void init_from_stack()
{
    from_stack_ptr = -1;
}






/**************************  push_from_fileblock
 **************************

Desc: pushes the given stream onto the the from file stack.
the stack pointer is preincremented and points at the
stacks top item.

In:   stream  file stream to be pushed.

 *****************************************************************
 ********/

void push_from_fileblock( a_from_fileblock )
    struct from_fileblock a_from_fileblock;
{
    from_file_stack[ ++from_stack_ptr ] = a_from_fileblock;
}



/**************************  pop_from_fileblock
 **************************

Desc: Pops the from file stack and returns the item that was on the
top.
The from stack pointer is post decremented and points to the
item
below the one popped.

Ret:  File stream popped from the top of the stack.

 *****************************************************************
 *******/

struct from_fileblock pop_from_fileblock()
{
    return( from_file_stack[ from_stack_ptr-- ] );
}




/*************************** trash_rest_of_line
 ***********************

Desc:  Sets >IN to point to the end of the current line in tib.

 *****************************************************************
 *******/

void trash_rest_of_line( void )
{
    unsigned CELL *lineptr;

    lineptr = (unsigned CELL *) &ram[ TO_IN ];
    *lineptr = strlen( &ram[ TIB ] );
}



/**************************  get_string
 *****************************

Desc: Attempts to get string delimited by the given character from
the
input stream. If one is found it is returned aCELL with the
value True. Otherwise False is returned.
If, during the composition of a delimited string the end of
line mark is encountered the end delimited string is
terminated
at that point (TRUE is still returned).

In:   delimiter  character which is the string delimeter

Out:  string     string returned if one found.

Ret:  True if delimited string was found, else False.


 *****************************************************************
 ****/

int get_string( string, delimiter )
    char string[], delimiter;
{
    int j=0;
    char *line;
    unsigned CELL *lineptr;

    lineptr = (unsigned CELL *) &ram[ TO_IN ];
    line = (char *) &ram[ TIB ];

    while( *(line + *lineptr) == delimiter ) (*lineptr)++;

    while( *(line + *lineptr) != '\0' &&
            *(line + *lineptr) != delimiter )
        string[ j++ ] = *(line + (*lineptr)++);

    string[ j ] = '\0';
    if( *(line + *lineptr) == delimiter ) (*lineptr)++;
    return( j != 0 );
}


/**************************  get_word
 *******************************

Desc: Pinc internal version of BL WORD

Out:  Forth word, number or junk that was delimited by spaces.

Ret:  True if word found else false.

 *****************************************************************
 ****/

int get_word( word )
    char word[];
{
    return( get_string( word, ' '));
}


/***************************  get_line
 *******************************

Desc: Gets a line of input form the file stream that is on top of
the from file stack.
If the top file stream is exhausted then each stream is
popped until either one is found unexhausted or stdin is
reached.

Ret:  Pointer to the line read in or NULL if none was.

Alg:  While top file stream exhausted and from stack at last item
pop from file stack
line read in or NULL is then returned.

 *****************************************************************
 ****/

int get_line( void )
{
    char *line;
    unsigned CELL *hash_tib, *lineptr;
    static int page_break_pending = 0;
    int status;

    lineptr = (unsigned CELL *) &ram[ TO_IN ];
    *lineptr = 0;
    line = (char *) &ram[ TIB ];
    hash_tib = (unsigned CELL *) &ram[ HASH_TIB ];
    *hash_tib = (CELL) 0;

    if( page_break_pending ) {
        input.page_number++;
        input.line_in_page_number = 0;
        page_break_pending = 0;
        while( input.page_number > input.last_page_required) {
            fprintf( output.stream, "done!\n" );
            fclose( input.stream );
            input = pop_from_fileblock();
            if( input.stream == stdin ) prompt();
        }
    }

    while( ((status = get_basic_line( line, LINE_LENGTH, input.stream, &page_break_pending))==EOF) && (from_stack_ptr != -1 )) {
        fprintf( output.stream, "Done!\n" );
        fclose( input.stream );
        input = pop_from_fileblock();
        if( input.stream == stdin ) prompt();
    }

    input.line_number++;
    input.line_in_page_number++;
    *hash_tib = (CELL) strlen( line );

    return( status );

}



/**************************  get_basic_line
 *****************************

Desc: Reads in a strgin from the given stream up to a max of length
characters long. Strings are delimited by \n, EOF and
page breaks (x0C). This is a special version of the fgets
function to cope with page breaks.

In:  lengththe max length of the line which can be read
stream    the stream from which we are to read

In/Out: line    line of text read in (possibly empty).
page_break  true if a page break terminated this string

Rets:Length of the string actually read in.

 *****************************************************************
 *****/

int get_basic_line( line, length, stream, page_break )
    char line[];
    int length, *page_break;
    FILE *stream;
{

    int i=0, ch=getc( stream );

//    printf("get_basic_line\n");

    while( (ch!=EOF) && (ch!='\x0C') && (ch!='\n') && (i<length)) {
        line[i++] = (( ch < 32) ? (char) 32 : (char) ch );
        ch=getc( stream );
    }

    *page_break = (ch == '\x0C');
    line[i] = '\0';
    return( ( (ch==EOF) && (i==0) ? EOF : !EOF ) );
}







/*********************************  prompt
 ******************************

Desc: If we are reading from standard in then the user is prompted
with ok and the data stack depth.

 *****************************************************************
 **********/

void prompt( void )
{
    if( from_stack_ptr == -1 ) {
        fprintf( output.stream, " ok" );
        if( sp != -1 )
#ifdef CELL_IS_LONG_LONG
        fprintf( output.stream, "-%lld", sp+1 );
#else
#ifdef CELL_IS_LONG
        fprintf( output.stream, "-%ld", sp+1 );
#else
        fprintf( output.stream, "-%d", sp+1 );
#endif
#endif
        fprintf( output.stream, "\n" );
    }
}




// #if IMPLEMENTATION == SUN
/*******************************  init_tty_config
 **************************

Desc: Initialises the tty configuration , saving it a away for
restoration after a raw setting or when a signal occurs.
This is done bye setting the terminal into raw mode, which
save the config before doing so, and then setting the tty
back into cooked. However this should only be
performed once, since when init_io is called from pinc
during a cold, say, we could be in either cooked or raw mode.
If we were in raw modewe would end up saving a raw
configuration `
rather than a cooked- very undesirable. We therefore ensure
this
code is only executed once, with ther done_once flag.

 *****************************************************************
 ***********/

init_tty_config()
{
    static int done_once=0;

    if( done_once ) return;
    done_once = 1;
    set_raw_io();
    set_cooked_io();
}



/*******************************  set_raw_io
 *******************************

Desc: Sets the given stream to be in raw mode if it is a terminal
otherwise nothing happens.

In:   stream   file stream to be set.

Alg:  We get the current tty configuration.
If this operation fails then the stream given is not a tty
and we go no further.
Otherwise we store of the terminal config in the global
variable
last_tty_config and set raw mode- acombination of non
canonical
input with a 0 number of minimum input characters, with no
wait
time. If the setting operation fails we report it.

 *****************************************************************
 *********/

void set_raw_io( void )
{
    /* ag-special. The following is commented out for compatibilty
       with SunOS 3.x. Can be uncommented for SUNOS 4.0
       to provide proper key? etc.
       */
    /*
       struct termio new_tty, previous_tty;

       if( ioctl( 00, TCGETA, &previous_tty ) != -1 ) {
       last_tty_config = previous_tty;
       new_tty = previous_tty;
       new_tty.c_lflag &= ~ICANON;
       new_tty.c_lflag &= ~ECHO;
       new_tty.c_cc[ VMIN ] = 0;
       new_tty.c_cc[ VTIME ] = 0;
       if( ioctl( 00, TCSETAF, &new_tty ) == -1 )
       fprintf( stderr, "Pinc: I/O Manager unable to make stream raw.\n" );
       }
       */
#if IMPLEMENTATION == Linux
    tcgetattr(0, &oldt);
    tcgetattr(0, &oldt);
    newt = oldt;
    newt.c_lflag &= ~(ICANON | ECHO);
    tcsetattr(0, TCSANOW, &newt);
    oldf = fcntl(0, F_GETFL, 0);
    //    fcntl(0, F_SETFL, oldf | O_NONBLOCK);
#endif

}



/*******************************  set_cooked_io
 *******************************

Desc: Sets the given stream to be in cooked mode if it is a
terminal
otherwise nothing happens.

Alg:  We get the current tty configuration.
If this operation fails then the stream given is not a tty
and we go no further.
Otherwise we restore of the last terminal config.

 *****************************************************************
 *********/

void set_cooked_io( void )
{
    /*  ag-special. The following is commented out for compatibilty
        with SunOS 3.x. Can be uncommented for SUNOS 4.0
        to provide proper key? etc.
        */
    /*
       struct termio previous_tty;

       if( ioctl( 00, TCGETA, &previous_tty ) != -1 )
       ioctl( 00, TCSETAF, &last_tty_config );
       */
    //
    // ATH This restores the settings that were saved by set_raw_io.
    // TODO set_raw_io is NOT idempotent.  Need to fix this.
    //
#if IMPLEMENTATION == Linux
    tcsetattr(0, TCSANOW, &oldt);
    fcntl(0, F_SETFL, oldf);
#endif
}


/*****************************  key_pressed
 ********************************

Desc: Returns true if a key was pressed else false. Stdin should
have
be set to raw via the set_raw_io function before this
function is called.

Ret:  True if key pressed else false.

 *****************************************************************
 ************/

int key_pressed( void ) {
    char ch;
    /*  ag-special. The following is commented out for compatibilty
        with SunOS 3.x. Can be uncommented for SUNOS 4.0
        to provide proper key? etc.
        */
    /*
       char buf[1];

       if( read( 00, buf, 1 ) >0 ) {
       ungetc( buf[0], stdin );
       return( 1 );
       }
       else
       return( 0 );
       */

#if IMPLEMENTATION == Linux

    fcntl(0, F_SETFL, oldf | O_NONBLOCK);

    ch = getchar();

    if(ch != EOF) {
        ungetc(ch, stdin);
        return 1;
    }

    return 0;

#endif
}
// #endif





/***************************  make_from_fileblock
 *************************

Desc: Makes a new from_fileblock with a given filename. The line
and page numbers are sert to 0 and the filestream is
undefined.

In:   filename string contatining the name of the file associated
with this block
stream filestream assocaited with this block

Rets: Returns a new from_fileblock

 *****************************************************************
 ***********/


struct from_fileblock make_from_fileblock( filename, stream )
    char *filename;
    FILE *stream;
{
    struct from_fileblock temp;

    strcpy( temp.filename, filename );
    temp.stream = stdin;
    temp.line_number = 0;
    temp.page_number = 1;
    temp.line_in_page_number = 0;
    temp.last_page_required = 9999;
    temp.stream = stream;
    return( temp );
}






/*****************************goto_page
 *****************************

Desc: Goes to the given page number in in the file if we have
not yet passed it. If given the page number exceeds the
number
of pages in the file then we read EOF and increment the
page counter until the page number is equal to the one we
want.

In:   Page_number  Page number to go to.

 *****************************************************************
 *****/

void goto_page( page_number )
    unsigned CELL page_number;
{
    char ch;

    while( input.page_number < (int) page_number ) {
        while( ((ch = getc( input.stream ) != '\x0C' )) && ( ch != EOF ));
        ++input.page_number;
    }
}
